#!python3
# -*- coding: utf-8 -*-
"""

"""
from typing import List, Dict, Tuple, Generator, Any, Union, OrderedDict
import io
import json
from warnings import warn

import numpy as np
import requests
from requests.auth import AuthBase
import pandas as pd

from ckanapi_harvesters.auxiliary.error_level_message import ContextErrorLevelMessage, ErrorLevel
from ckanapi_harvesters.auxiliary.list_records import ListRecords, records_to_df
from ckanapi_harvesters.auxiliary.proxy_config import ProxyConfig
from ckanapi_harvesters.auxiliary.ckan_model import CkanResourceInfo, CkanAliasInfo, CkanField
from ckanapi_harvesters.auxiliary.ckan_map import CkanMap
from ckanapi_harvesters.auxiliary.ckan_auxiliary import bytes_to_megabytes
from ckanapi_harvesters.auxiliary.ckan_auxiliary import assert_or_raise, CkanIdFieldTreatment
from ckanapi_harvesters.auxiliary.ckan_auxiliary import datastore_id_col
from ckanapi_harvesters.auxiliary.ckan_auxiliary import RequestType
from ckanapi_harvesters.auxiliary.ckan_action import CkanActionResponse, CkanNotFoundError, CkanSqlCapabilityError
from ckanapi_harvesters.auxiliary.ckan_errors import (IntegrityError, CkanServerError, CkanArgumentError, SearchAllNoCountsError,
                                                      DataStoreNotFoundError, RequestError)
from ckanapi_harvesters.ckan_api.ckan_api_params import CkanApiParamsBasic
from ckanapi_harvesters.auxiliary.ckan_api_key import CkanApiKey
from ckanapi_harvesters.ckan_api.ckan_api_0_base import ckan_request_proxy_default_auth_if_ckan

from ckanapi_harvesters.ckan_api.ckan_api_1_map import CkanApiMap

df_download_read_csv_kwargs = dict(keep_default_na=False)

ckan_dtype_mapper = {
    "text": "str",
    "numeric": "float",
    "timestamp": "datetime64",
    "int": "int",
    "name": "str",
    "oid": "str",  # to confirm
    "bool": "object",  # enable None values but if they are present, booleans are converted to str...
    "json": "object",
}

class CkanApiReadOnlyParams(CkanApiParamsBasic):
    map_all_aliases:bool = True
    default_df_download_id_field_treatment: CkanIdFieldTreatment = CkanIdFieldTreatment.SetIndex

    def __init__(self, *, proxies:Union[str,dict,ProxyConfig]=None,
                 ckan_headers:dict=None, http_headers:dict=None):
        super().__init__(proxies=proxies, ckan_headers=ckan_headers, http_headers=http_headers)
        self.df_download_id_field_treatment: CkanIdFieldTreatment = self.default_df_download_id_field_treatment

    def copy(self, new_identifier:str=None, *, dest=None):
        if dest is None:
            dest = CkanApiReadOnlyParams()
        super().copy(dest=dest)
        dest.df_download_id_field_treatment = self.df_download_id_field_treatment
        return dest


## Main class ------------------
class CkanApiReadOnly(CkanApiMap):
    """
    CKAN Database API interface to CKAN server with helper functions using pandas DataFrames.
    This class implements requests to read data from the CKAN server resources / DataStores.
    """

    def __init__(self, url:str=None, *, proxies:Union[str,dict,ProxyConfig]=None,
                 apikey:Union[str,CkanApiKey]=None, apikey_file:str=None,
                 owner_org:str=None, params:CkanApiReadOnlyParams=None,
                 map:CkanMap=None,
                 identifier=None):
        """
        CKAN Database API interface to CKAN server with helper functions using pandas DataFrames.

        :param url: url of the CKAN server
        :param proxies: proxies to use for requests
        :param apikey: way to provide the API key directly (optional)
        :param apikey_file: path to a file containing a valid API key in the first line of text (optional)
        :param owner_org: name of the organization to limit package_search (optional)
        :param params: other connection/behavior parameters
        :param map: map of known resources
        :param identifier: identifier of the ckan client
        """
        super().__init__(url=url, proxies=proxies, apikey=apikey, apikey_file=apikey_file,
                         owner_org=owner_org, map=map, identifier=identifier)
        if params is None:
            params = CkanApiReadOnlyParams()
        if proxies is not None:
            params.proxies = proxies
        self.params: CkanApiReadOnlyParams = params

    def _rx_records_df_clean(self, df: pd.DataFrame) -> None:
        """
        Auxiliary function for cleaning dataframe from DataStore requests

        :param df:
        :return:
        """
        if len(df) > 0 and datastore_id_col in df.columns:
            if self.params.df_download_id_field_treatment == CkanIdFieldTreatment.SetIndex:
                # use _id column as new index
                df.set_index(datastore_id_col, drop=False, inplace=True, verify_integrity=True)
            elif self.params.df_download_id_field_treatment == CkanIdFieldTreatment.Remove:
                # remove "_id" column
                df.pop(datastore_id_col)

    @staticmethod
    def read_fields_type_dict(fields_list_dict: List[dict]) -> OrderedDict:
        return OrderedDict([(field_dict["id"], field_dict["type"]) for field_dict in fields_list_dict])

    @staticmethod
    def read_fields_df_args(fields_type_dict: OrderedDict) -> dict:
        if fields_type_dict is None:
            return {}
        # fields_dtype_dict = fields_type_dict.copy()
        # for key, ckan_type in fields_type_dict.items():
        #     if ckan_type in ckan_dtype_mapper:
        #         fields_dtype_dict[key] = ckan_dtype_mapper[ckan_type]
        #     else:
        #         fields_dtype_dict[key] = "object"
        # return dict(names=list(fields_dtype_dict.keys()), dtype=fields_dtype_dict)
        return dict(names=list(fields_type_dict.keys()))

    @staticmethod
    def from_dict_df_args(fields_type_dict: OrderedDict) -> dict:
        df_args_dict = CkanApiReadOnly.read_fields_df_args(fields_type_dict)
        df_args_dict.pop("names")
        return df_args_dict

    ## Data queries ------------------
    ### Dump method ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
    # NB: dump methods are not exposed to the user by default. Only datastore_search and resource_download methods are exposed.
    def _api_datastore_dump_raw(self, resource_id:str, *, filters:dict=None, q:str=None, fields:List[str]=None,
                                sort:str=None, limit:int=None, offset:int=0, format:str=None, bom:bool=None, params:dict=None,
                                compute_len:bool=False) -> requests.Response:
        """
        URL call to datastore/dump URL. Dumps successive lines in the DataStore.

        :param resource_id: resource id.
        :param filters: The base argument to filter values in a table (optional)
        :param q: Full text query (optional)
        :param fields: The base argument to filter columns (optional)
        :param format: The return format in the returned response (default=csv, tsv, json, xml) (optional)
        :param params: Additional parameters such as filters, q, sort and fields can be given. See DataStore API documentation.
        :return: raw response
        """
        if compute_len:
            raise SearchAllNoCountsError("datastore_search", f"format={format}")
        if params is None:
            params = {}
        if offset is None:
            offset = 0
        params["offset"] = offset
        if limit is None:
            limit = self.params.default_limit_read
        if limit is not None:
            params["limit"] = limit
        if filters is not None:
            if isinstance(filters, str):
                # not recommended
                params["filters"] = filters
            else:
                params["filters"] = json.dumps(filters)
        if q is not None:
            params["q"] = q
        if fields is not None:
            params["fields"] = fields
        if sort is not None:
            params["sort"] = sort
        if format is not None:
            format = format.lower()
            params["format"] = format
        if bom is None and format is not None:
            bom = format not in {"json", "xml"}
        if bom is not None:
            params["bom"] = bom
        # params["bom"] = True  # useful?
        response = self._url_request(f"datastore/dump/{resource_id}", method=RequestType.Get, params=params)
        if response.status_code == 200:
            return response
        elif response.status_code == 404 and "DataStore resource not found" in response.text:
            raise DataStoreNotFoundError(resource_id, response.content.decode())
        else:
            raise CkanServerError(self, response, response.content.decode())

    def _api_datastore_dump_df(self, resource_id:str, *, filters:dict=None, q:str=None, fields:List[str]=None,
                               sort:str=None, limit:int=None, offset:int=0, format:str=None, bom:bool=None, params:dict=None) -> pd.DataFrame:
        """
        Convert output of _api_datastore_dump_raw to pandas DataFrame.
        """
        response = self._api_datastore_dump_raw(resource_id=resource_id, filters=filters, q=q, fields=fields,
                                                sort=sort, limit=limit, offset=offset, format=format, bom=bom,
                                                params=params, compute_len=False)
        if format is not None:
            format = format.lower()
        buffer = io.StringIO(response.content.decode())
        if format is None or format == "csv":
            response_df = pd.read_csv(buffer, **df_download_read_csv_kwargs)
        elif format == "tsv":
            response_df = pd.read_csv(buffer, sep="\t", **df_download_read_csv_kwargs)  # not tested
        elif format == "json":
            response_dict = json.load(buffer)
            fields_type_dict = CkanApiReadOnly.read_fields_type_dict(response_dict["fields"])
            df_args = CkanApiReadOnly.read_fields_df_args(fields_type_dict)
            response_df = records_to_df(response_dict["records"], df_args)
            response_df.attrs["fields"] = fields_type_dict
        elif format == "xml":
            response_df = pd.read_xml(buffer, parser="etree") # , xpath=".//row")  # partially tested  # otherwise, necessitates the installation of parser lxml
        else:
            raise NotImplementedError()
        self._rx_records_df_clean(response_df)
        return response_df

    def _api_datastore_dump_all(self, resource_id:str, *, filters:dict=None, q:str=None, fields:List[str]=None,
                                sort:str=None, limit:int=None, offset:int=0, format:str=None, bom:bool=None,
                                params:dict=None, search_all:bool=True, return_df:bool=True) \
            -> Union[pd.DataFrame, requests.Response]:
        """
        Successive calls to _api_datastore_dump_df until an empty list is received.

        :see: _api_datastore_dump()
        :param resource_id: resource id.
        :param filters: The base argument to filter values in a table (optional)
        :param q: Full text query (optional)
        :param fields: The base argument to filter columns (optional)
        :param format: The return format in the returned response (default=csv, tsv, json, xml) (optional)
        :param params: Additional parameters such as filters, q, sort and fields can be given. See DataStore API documentation.
        :param search_all: if False, only the first request is operated
        :return:
        """
        if return_df:
            return self._request_all_results_df(api_fun=self._api_datastore_dump_df, params=params, limit=limit, offset=offset,
                                                search_all=search_all, resource_id=resource_id,
                                                filters=filters, q=q, fields=fields, sort=sort, format=format, bom=bom)
        elif search_all:
            # cannot determine the number of records received if the response is not parsed with pandas in this mode
            # at least, the total number of rows should be known
            # concatenation of results requires parsing of the result
            # => this mode is useless => raise error
            raise SearchAllNoCountsError("datastore_dump")
        else:
            response = self._api_datastore_dump_raw(resource_id=resource_id, filters=filters, q=q, fields=fields,
                                                    sort=sort, limit=limit, offset=offset, format=format, bom=bom,
                                                    params=params, compute_len=search_all)
            return response

    def _api_datastore_dump_all_generator(self, resource_id:str, *, filters:dict=None, q:str=None, fields:List[str]=None,
                                          sort:str=None, limit:int=None, offset:int=0, format:str=None, bom:bool=None,
                                          params:dict=None, search_all:bool=True, return_df:bool=True) \
            -> Union[Generator[pd.DataFrame, Any, None], Generator[requests.Response, Any, None]]:
        """
        Successive calls to _api_datastore_dump until an empty list is received.
        Generator implementation which yields one DataFrame per request.

        :see: _api_datastore_dump()
        :param resource_id: resource id.
        :param filters: The base argument to filter values in a table (optional)
        :param q: Full text query (optional)
        :param fields: The base argument to filter columns (optional)
        :param format: The return format in the returned response (default=csv, tsv, json, xml) (optional)
        :param params: Additional parameters such as filters, q, sort and fields can be given. See DataStore API documentation.
        :param search_all: if False, only the first request is operated
        :return:
        """
        if return_df:
            return self._request_all_results_generator(api_fun=self._api_datastore_dump_df, params=params, limit=limit, offset=offset,
                                                       search_all=search_all, resource_id=resource_id,
                                                       filters=filters, q=q, fields=fields, sort=sort, format=format, bom=bom)
        else:
            return self._request_all_results_generator(api_fun=self._api_datastore_dump_raw, params=params, limit=limit, offset=offset,
                                                       search_all=search_all, resource_id=resource_id,
                                                       filters=filters, q=q, fields=fields, sort=sort, format=format, bom=bom,
                                                       compute_len=search_all)


    ### Search method ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
    def _api_datastore_search_raw(self, resource_id:str, *, filters:dict=None, q:str=None, fields:List[str]=None,
                                  distinct:bool=None, sort:str=None, limit:int=None, offset:int=0, format:str=None,
                                  params:dict=None, compute_len:int=False) -> CkanActionResponse:
        """
        API call to datastore_search. Performs queries on the DataStore.

        :param resource_id: resource id.
        :param filters: The base argument to filter values in a table (optional)
        :param q: Full text query (optional)
        :param fields: The base argument to filter columns (optional)
        :param distinct: return only distinct rows (optional, default: false) e.g. to return distinct ids: fields="id", distinct=True
        :param sort: Argument to sort results e.g. sort="index, quantity desc"  or  sort="index asc"
        :param limit: Limit the number of records to return
        :param offset: Offset in the returned records
        :param format: The return format in the returned response (default=objects, csv, tsv, lists) (optional)
        :param params: Additional parameters such as filters, q, sort and fields can be given. See DataStore API documentation.
        :return:
        """
        if params is None:
            params = {}
        if offset is None:
            offset = 0
        params["offset"] = offset
        if limit is None:
            limit = self.params.default_limit_read
        if limit is not None:
            params["limit"] = limit
        params["resource_id"] = resource_id
        if filters is not None:
            if isinstance(filters, str):
                # not recommended
                params["filters"] = filters
            else:
                params["filters"] = json.dumps(filters)
        if q is not None:
            params["q"] = q
        if fields is not None:
            params["fields"] = fields
        if distinct is not None:
            params["distinct"] = distinct
        if sort is not None:
            params["sort"] = sort
        if format is not None:
            format = format.lower()
            params["records_format"] = format
        response = self._api_action_request(f"datastore_search", method=RequestType.Get, params=params)
        if response.success:
            if response.dry_run:
                return response
            elif format is None or format in ["objects", "lists"]:
                response.len = len(response.result["records"])
            elif compute_len:
                raise SearchAllNoCountsError("datastore_search", f"format={format}")
            return response
        elif response.status_code == 404 and response.error_message["__type"] == "Not Found Error":
            raise DataStoreNotFoundError(resource_id, response.error_message)
        else:
            raise response.default_error(self)

    def _api_datastore_search_df(self, resource_id:str, *, filters:dict=None, q:str=None, fields:List[str]=None,
                                 distinct:bool=None, sort:str=None, limit:int=None, offset:int=0, format:str=None, params:dict=None) -> pd.DataFrame:
        """
        Convert output of _api_datastore_search_raw to pandas DataFrame.
        """
        response = self._api_datastore_search_raw(resource_id=resource_id, filters=filters, q=q, fields=fields, format=format,
                                                  distinct=distinct, sort=sort, limit=limit, offset=offset,
                                                  params=params, compute_len=False)
        if response.dry_run:
            return pd.DataFrame()
        if format is not None:
            format = format.lower()
        fields_type_dict = CkanApiReadOnly.read_fields_type_dict(response.result["fields"])
        if format is None or format == "objects":
            df_args_dict = CkanApiReadOnly.from_dict_df_args(fields_type_dict)
            response_df = pd.DataFrame.from_dict(response.result["records"], **df_args_dict)
        else:
            df_args = CkanApiReadOnly.read_fields_df_args(fields_type_dict)
            if format == "lists":
                response_df = records_to_df(response.result["records"], df_args)
            else:
                buffer = io.StringIO(response.result["records"])
                if format == "csv":
                    response_df = pd.read_csv(buffer, **df_args, **df_download_read_csv_kwargs)
                elif format == "tsv":
                    response_df = pd.read_csv(buffer, sep='\t', **df_args, **df_download_read_csv_kwargs)
                else:
                    raise NotImplementedError()
        self._rx_records_df_clean(response_df)
        response.result.pop("records")
        response_df.attrs["result"] = response.result
        response_df.attrs["fields"] = fields_type_dict
        response_df.attrs["total"] = response.result["total"]
        response_df.attrs["total_was_estimated"] = response.result["total_was_estimated"]
        response_df.attrs["limit"] = response.result["limit"]
        return response_df

    def _api_datastore_search_all(self, resource_id:str, *, filters:dict=None, q:str=None, fields:List[str]=None,
                                  distinct:bool=None, sort:str=None, limit:int=None, offset:int=0, format:str=None,
                                  search_all:bool=True, params:dict=None, return_df:bool=True, compute_len:bool=False) \
            -> Union[pd.DataFrame, Tuple[ListRecords, OrderedDict], Any]:
        """
        Successive calls to _api_datastore_search_df until an empty list is received.

        :see: _api_datastore_search()
        :param resource_id: resource id.
        :param filters: The base argument to filter values in a table (optional)
        :param q: Full text query (optional)
        :param fields: The base argument to filter columns (optional)
        :param distinct: return only distinct rows (optional, default: false) e.g. to return distinct ids: fields="id", distinct=True
        :param sort: Argument to sort results e.g. sort="index, quantity desc"  or  sort="index asc"
        :param limit: Limit the number of records to return
        :param offset: Offset in the returned records
        :param format: The return format in the returned response (default=objects, csv, tsv, lists) (optional)
        :param params: Additional parameters such as filters, q, sort and fields can be given. See DataStore API documentation.
        :param search_all: if False, only the first request is operated
        :return:
        """
        if return_df:
            df = self._request_all_results_df(api_fun=self._api_datastore_search_df, params=params, limit=limit, offset=offset,
                                              search_all=search_all, resource_id=resource_id, filters=filters, q=q, fields=fields, distinct=distinct, sort=sort, format=format)
            if "fields" in df.attrs.keys():
                df.attrs["fields"] = df.attrs["fields"][0]
            if "total" in df.attrs.keys():
                assert_or_raise(np.all(np.array(df.attrs["total"]) == df.attrs["total"][0]), IntegrityError("total field varied in the responses"))
                df.attrs["total"] = df.attrs["total"][0]
            return df
        else:
            responses = self._request_all_results_list(api_fun=self._api_datastore_search_raw, params=params, limit=limit, offset=offset,
                                            search_all=search_all, resource_id=resource_id, filters=filters, q=q, fields=fields, distinct=distinct, sort=sort, format=format, compute_len=compute_len)
            # aggregate results, depending on the format
            if self.params.dry_run:
                return [], {}
            if format is not None:
                format = format.lower()
            if len(responses) > 0:
                response = responses[0]
                fields_type_dict = CkanApiReadOnly.read_fields_type_dict(response.result["fields"])
                df_args = CkanApiReadOnly.read_fields_df_args(fields_type_dict)
            else:
                fields_type_dict = None
                df_args = {}
            if format is None or format == "objects":
                return ListRecords(sum([response.result["records"] for response in responses], [])), fields_type_dict
            else:
                if format == "lists":
                    return sum([response.result["records"] for response in responses], []), fields_type_dict
                else:
                    return "\n".join([response.result["records"] for response in responses]), fields_type_dict

    def _api_datastore_search_all_generator(self, resource_id:str, *, filters:dict=None, q:str=None, fields:List[str]=None,
                                            distinct:bool=None, sort:str=None, limit:int=None, offset:int=0,
                                            format:str=None, search_all:bool=True, params:dict=None, return_df:bool=True) \
            -> Union[Generator[pd.DataFrame, Any, None], Generator[CkanActionResponse, Any, None]]:
        """
        Successive calls to _api_datastore_search_df until an empty list is received.
        Generator implementation which yields one DataFrame per request.

        :see: _api_datastore_search()
        :param resource_id: resource id.
        :param filters: The base argument to filter values in a table (optional)
        :param q: Full text query (optional)
        :param fields: The base argument to filter columns (optional)
        :param distinct: return only distinct rows (optional, default: false) e.g. to return distinct ids: fields="id", distinct=True
        :param sort: Argument to sort results e.g. sort="index, quantity desc"  or  sort="index asc"
        :param limit: Limit the number of records to return
        :param offset: Offset in the returned records
        :param format: The return format in the returned response (default=objects, csv, tsv, lists) (optional)
        :param params: Additional parameters such as filters, q, sort and fields can be given. See DataStore API documentation.
        :param search_all: if False, only the first request is operated
        :return:
        """
        if return_df:
            return self._request_all_results_generator(api_fun=self._api_datastore_search_df, params=params, limit=limit, offset=offset,
                                                       search_all=search_all, resource_id=resource_id, filters=filters, q=q, fields=fields, distinct=distinct, sort=sort, format=format, compute_len=True)
        else:
            return self._request_all_results_generator(api_fun=self._api_datastore_search_raw, params=params,
                                                       limit=limit, offset=offset, search_all=search_all,
                                                       resource_id=resource_id, filters=filters, q=q,
                                                       fields=fields, distinct=distinct, sort=sort,
                                                       format=format, compute_len=search_all)


    ### search_sql method ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
    def _api_datastore_search_sql_raw(self, sql:str, *, params:dict=None, limit:int=None, offset:int=0) -> CkanActionResponse:
        """
        API call to datastore_search_sql. Performs SQL queries on the DataStore. These queries can be more complex than
        with datastore_search. The DataStores are referenced by their resource_id, surrounded by quotes. The field names
        are referred by their name in upper case, surrounded by quotes.
        __NB__: This action is not available when ckanapi_harvesters.datastore.sqlsearch.enabled is set to false

        :param sql: SQL query e.g. f'SELECT * IN "{resource_id}" WHERE "USER_ID" < 0'
        :param limit: Limit the number of records to return
        :param offset: Offset in the returned records
        :param params: N/A
        :return:
        """
        if params is None:
            params = {}
        params["sql"] = sql
        if offset is None:
            offset = 0
        params["offset"] = offset
        if limit is None:
            limit = self.params.default_limit_read
        if limit is not None:
            params["limit"] = limit
        response = self._api_action_request(f"datastore_search_sql", method=RequestType.Post, params=params)
        if response.success:
            return response
        elif response.status_code == 400 and response.success_json_loads and response.response.text == '"Bad request - Action name not known: datastore_search_sql"':
            raise CkanSqlCapabilityError(self, response)
        elif response.status_code == 404 and response.success_json_loads and response.error_message["__type"] == "Not Found Error":
            raise CkanNotFoundError(self, "SQL", response)
        else:
            raise response.default_error(self)

    def _api_datastore_search_sql_df(self, sql:str, *, params:dict=None, limit:int=None, offset:int=0) -> pd.DataFrame:
        """
        Convert output of _api_datastore_search_sql_raw to pandas DataFrame.
        """
        response = self._api_datastore_search_sql_raw(sql=sql, params=params, limit=limit, offset=offset)
        fields_type_dict = CkanApiReadOnly.read_fields_type_dict(response.result["fields"])
        df_args_dict = CkanApiReadOnly.from_dict_df_args(fields_type_dict)
        response_df = pd.DataFrame.from_dict(response.result["records"], **df_args_dict)
        response.result.pop("records")
        response_df.attrs["result"] = response.result
        response_df.attrs["fields"] = fields_type_dict
        # response_df.attrs["total"] = response.result["total"]
        # response_df.attrs["total_was_estimated"] = response.result["total_was_estimated"]
        response_df.attrs["limit"] = response.result["limit"]
        self._rx_records_df_clean(response_df)
        return response_df

    def _api_datastore_search_sql_all(self, sql:str, *, params:dict=None,
                                      search_all:bool=True, limit:int=None, offset:int=0, return_df:bool=True) \
            -> Union[pd.DataFrame, Tuple[ListRecords, dict]]:
        """
        Successive calls to _api_datastore_search_sql until an empty list is received.

        :see: _api_datastore_search_sql()
        :param sql: SQL query e.g. f'SELECT * IN "{resource_id}" WHERE "USER_ID" < 0'
        :param limit: Limit the number of records to return
        :param offset: Offset in the returned records
        :param params: N/A
        :param search_all: if False, only the first request is operated
        :return:
        """
        if return_df:
            df = self._request_all_results_df(api_fun=self._api_datastore_search_sql_df, params=params,
                                              limit=limit, offset=offset, search_all=search_all, sql=sql)
            if "fields" in df.attrs.keys():
                df.attrs["fields"] = df.attrs["fields"][0]
            # if "total" in df.attrs.keys():
            #     assert_or_raise(np.all(np.array(df.attrs["total"]) == df.attrs["total"][0]), IntegrityError("total field varied in the responses"))
            #     df.attrs["total"] = df.attrs["total"][0]
            return df
        else:
            responses = self._request_all_results_list(api_fun=self._api_datastore_search_sql_raw, params=params,
                                                       limit=limit, offset=offset, search_all=search_all, sql=sql)
            # TODO: test
            if len(responses) > 0:
                response = responses[0]
                fields_type_dict = CkanApiReadOnly.read_fields_type_dict(response.result["fields"])
            else:
                fields_type_dict = None
            return ListRecords(sum([response.result["records"] for response in responses], [])), fields_type_dict

    def _api_datastore_search_sql_all_generator(self, sql:str, *, params:dict=None,
                                                search_all:bool=True, limit:int=None, offset:int=0, return_df:bool=True) \
            -> Union[Generator[pd.DataFrame, Any, None], Generator[CkanActionResponse, Any, None]]:
        """
        Successive calls to _api_datastore_search_sql until an empty list is received.
        Generator implementation which yields one DataFrame per request.

        :see: _api_datastore_search_sql()
        :param sql: SQL query e.g. f'SELECT * IN "{resource_id}" WHERE "USER_ID" < 0'
        :param limit: Limit the number of records to return
        :param offset: Offset in the returned records
        :param params: N/A
        :param search_all: if False, only the first request is operated
        :return:
        """
        if return_df:
            return self._request_all_results_generator(api_fun=self._api_datastore_search_sql_df, params=params,
                                                       limit=limit, offset=offset, search_all=search_all, sql=sql)
        else:
            return self._request_all_results_generator(api_fun=self._api_datastore_search_sql_raw, params=params,
                                                       limit=limit, offset=offset, search_all=search_all, sql=sql)


    ## Function aliases to limit the entry-points for the user  -------------------------------------------------------
    def datastore_search(self, resource_id:str, *, filters:dict=None, q:str=None, fields:List[str]=None,
                         distinct:bool=None, sort:str=None, limit:int=None, offset:int=0, params:dict=None,
                         search_all:bool=False, search_method:bool=True, format:str=None, return_df:bool=True) \
            -> Union[pd.DataFrame, ListRecords, Any, List[CkanActionResponse]]:
        """
        Preferred entry-point for a DataStore read request.
        Uses the API datastore_search

        :param resource_id: resource id.
        :param filters: The base argument to filter values in a table (optional)
        :param q: Full text query (optional)
        :param fields: The base argument to filter columns (optional)
        :param distinct: return only distinct rows (optional, default: false) e.g. to return distinct ids: fields="id", distinct=True
        :param sort: Argument to sort results e.g. sort="index, quantity desc"  or  sort="index asc"
        :param limit: Limit the number of records to return
        :param offset: Offset in the returned records
        :param params: Additional parameters such as filters, q, sort and fields can be given. See DataStore API documentation.
        :param search_all: Option to renew the request until there are no more records.
        :param search_method: API method selection (True=datastore_search, False=datastore_dump)
        :return:
        """
        if search_method:
            if return_df and format is None: format = "csv"
            return self._api_datastore_search_all(resource_id, filters=filters, q=q, fields=fields, distinct=distinct, sort=sort,
                                                  limit=limit, offset=offset, format=format, params=params, search_all=search_all, return_df=return_df)
        else:
            assert_or_raise(distinct is None, CkanArgumentError("DataStore dump", "distinct"))
            if return_df and format is None: format, bom = "csv", True
            return self._api_datastore_dump_all(resource_id, filters=filters, q=q, fields=fields, sort=sort,
                                                limit=limit, offset=offset, format=format, bom=bom, params=params, search_all=search_all, return_df=return_df)

    def datastore_dump(self, resource_id:str, *, filters:dict=None, q:str=None, fields:List[str]=None,
                         distinct:bool=None, sort:str=None, limit:int=None, offset:int=0, params:dict=None,
                         search_all:bool=True, search_method:bool=True, format:str=None, return_df:bool=True) \
            -> Union[pd.DataFrame, ListRecords, Any, List[CkanActionResponse]]:
        """
        Alias of datastore_search with search_all=True by default.
        Uses the API datastore_search

        :see: datastore_search()
        :param resource_id: resource id.
        :param filters: The base argument to filter values in a table (optional)
        :param q: Full text query (optional)
        :param fields: The base argument to filter columns (optional)
        :param distinct: return only distinct rows (optional, default: false) e.g. to return distinct ids: fields="id", distinct=True
        :param sort: Argument to sort results e.g. sort="index, quantity desc"  or  sort="index asc"
        :param limit: Limit the number of records to return
        :param offset: Offset in the returned records
        :param params: Additional parameters such as filters, q, sort and fields can be given. See DataStore API documentation.
        :param search_all: Option to renew the request until there are no more records.
        :param search_method: API method selection (True=datastore_search, False=datastore_dump)
        :return:
        """
        return self.datastore_search(resource_id, filters=filters, q=q, fields=fields,
                                     distinct=distinct, sort=sort, limit=limit, offset=offset, params=params,
                                     search_all=search_all, search_method=search_method, format=format, return_df=return_df)

    def datastore_search_generator(self, resource_id:str, *, filters:dict=None, q:str=None, fields:List[str]=None,
                                   distinct:bool=None, sort:str=None, limit:int=None, offset:int=0, params:dict=None,
                                   search_all:bool=False, search_method:bool=True, format:str=None, return_df:bool=True) \
            -> Union[Generator[pd.DataFrame, Any, None], Generator[CkanActionResponse, Any, None], Generator[requests.Response, Any, None]]:
        """
        Preferred entry-point for a DataStore read request.
        Uses the API datastore_search

        :param resource_id: resource id.
        :param filters: The base argument to filter values in a table (optional)
        :param q: Full text query (optional)
        :param fields: The base argument to filter columns (optional)
        :param distinct: return only distinct rows (optional, default: false) e.g. to return distinct ids: fields="id", distinct=True
        :param sort: Argument to sort results e.g. sort="index, quantity desc"  or  sort="index asc"
        :param limit: Limit the number of records to return
        :param offset: Offset in the returned records
        :param params: Additional parameters such as filters, q, sort and fields can be given. See DataStore API documentation.
        :param search_all: Option to renew the request until there are no more records.
        :param search_method: API method selection (True=datastore_search, False=datastore_dump)
        :return:
        """
        if search_method:
            if return_df and format is None: format = "csv"
            return self._api_datastore_search_all_generator(resource_id, filters=filters, q=q, fields=fields, distinct=distinct, sort=sort,
                                                               limit=limit, offset=offset, format=format, params=params, search_all=search_all, return_df=return_df)
        else:
            assert_or_raise(distinct is None, CkanArgumentError("DataStore dump", "distinct"))
            if return_df and format is None: format, bom = "csv", True
            return self._api_datastore_dump_all_generator(resource_id, filters=filters, q=q, fields=fields, sort=sort,
                                                          limit=limit, offset=offset, format=format, bom=bom, params=params, search_all=search_all, return_df=return_df)

    def datastore_search_cursor(self, resource_id:str, *, filters:dict=None, q:str=None, fields:List[str]=None,
                                   distinct:bool=None, sort:str=None, limit:int=None, offset:int=0, params:dict=None,
                                   search_all:bool=False, search_method:bool=True, format:str=None, return_df:bool=True) \
            -> Generator[Union[pd.Series, Tuple[dict,dict], Tuple[list,dict], Tuple[str,dict]], Any, None]:
        """
        Cursor on rows
        """
        generator = self.datastore_search_generator(resource_id, filters=filters, q=q, fields=fields,
                                                    distinct=distinct, sort=sort, limit=limit, offset=offset, params=params,
                                                    search_all=search_all, search_method=search_method, format=format, return_df=return_df)
        if return_df:
            df: pd.DataFrame
            row: pd.Series
            for df in generator:
                for index, row in df.iterrows():
                    yield row
        elif search_method:
            response: CkanActionResponse
            # response.result: list
            if format is not None:
                format = format.lower()
            if format is None or format == "objects":
                for response in generator:
                    fields_type_dict = CkanApiReadOnly.read_fields_type_dict(response.result["fields"])
                    for element in response.result["records"]:
                        yield element, fields_type_dict
            else:
                for response in generator:
                    fields_type_dict = CkanApiReadOnly.read_fields_type_dict(response.result["fields"])
                    for element in response.result["records"]:
                        yield element, fields_type_dict
        else:
            raise TypeError("dumping datastore without parsing with a DataFrame does not return an iterable object")

    def datastore_dump_generator(self, resource_id:str, *, filters:dict=None, q:str=None, fields:List[str]=None,
                                   distinct:bool=None, sort:str=None, limit:int=None, offset:int=0, params:dict=None,
                                   search_all:bool=True, search_method:bool=True, format:str=None, return_df:bool=True) \
            -> Union[Generator[pd.DataFrame, Any, None], Generator[CkanActionResponse, Any, None]]:
        """
        Function alias to datastore_search_generator with search_all=True by default.
        Uses the API datastore_search

        :see: datastore_search_generator
        :param resource_id: resource id.
        :param filters: The base argument to filter values in a table (optional)
        :param q: Full text query (optional)
        :param fields: The base argument to filter columns (optional)
        :param distinct: return only distinct rows (optional, default: false) e.g. to return distinct ids: fields="id", distinct=True
        :param sort: Argument to sort results e.g. sort="index, quantity desc"  or  sort="index asc"
        :param limit: Limit the number of records to return
        :param offset: Offset in the returned records
        :param params: Additional parameters such as filters, q, sort and fields can be given. See DataStore API documentation.
        :param search_all: Option to renew the request until there are no more records.
        :param search_method: API method selection (True=datastore_search, False=datastore_dump)
        :return:
        """
        return self.datastore_search_generator(resource_id, filters=filters, q=q, fields=fields,
                                               distinct=distinct, sort=sort, limit=limit, offset=offset, params=params,
                                               search_all=search_all, search_method=search_method, format=format, return_df=return_df)

    def datastore_search_sql(self, sql:str, *, params:dict=None, search_all:bool=False,
                             limit:int=None, offset:int=0, return_df:bool=True) -> Union[pd.DataFrame, Tuple[ListRecords, dict]]:
        """
        Preferred entry-point for a DataStore SQL request.
        :see: _api_datastore_search_sql()
        __NB__: This action is not available when ckanapi_harvesters.datastore.sqlsearch.enabled is set to false

        :param sql: SQL query e.g. f'SELECT * IN "{resource_id}" WHERE "USER_ID" < 0'
        :param limit: Limit the number of records to return
        :param offset: Offset in the returned records
        :param params: N/A
        :param search_all: Option to renew the request until there are no more records.
        :return:
        """
        return self._api_datastore_search_sql_all(sql, params=params, limit=limit, offset=offset, search_all=search_all, return_df=return_df)

    def datastore_search_sql_generator(self, sql:str, *, params:dict=None, search_all:bool=False,
                                       limit:int=None, offset:int=0, return_df:bool=True) \
            -> Union[Generator[pd.DataFrame, Any, None], Generator[CkanActionResponse, Any, None]]:
        """
        Preferred entry-point for a DataStore SQL request.
        :see: _api_datastore_search_sql()

        __NB__: This action is not available when ckanapi_harvesters.datastore.sqlsearch.enabled is set to false

        :param sql: SQL query e.g. f'SELECT * IN "{resource_id}" WHERE "USER_ID" < 0'
        :param limit: Limit the number of records to return
        :param offset: Offset in the returned records
        :param params: N/A
        :param search_all: Option to renew the request until there are no more records.
        :return:
        """
        return self._api_datastore_search_sql_all_generator(sql, params=params, limit=limit, offset=offset, search_all=search_all, return_df=return_df)

    def datastore_search_sql_cursor(self, sql:str, *, params:dict=None, search_all:bool=False,
                                    limit:int=None, offset:int=0, return_df:bool=True) \
            -> Generator[Union[pd.Series,Tuple[dict,dict]], Any, None]:
        generator = self.datastore_search_sql_generator(sql, params=params, search_all=search_all,
                                                        limit=limit, offset=offset, return_df=return_df)
        if return_df:
            df: pd.DataFrame
            row: pd.Series
            for df in generator:
                for index, row in df.iterrows():
                    yield row
        else:
            response: CkanActionResponse
            # response.result: list
            element: Any
            for response in generator:
                fields_type_dict = CkanApiReadOnly.read_fields_type_dict(response.result["fields"])
                for element in response.result["records"]:
                    yield element, fields_type_dict

    def datastore_search_sql_find_one(self, sql:str, *, params:dict=None,
                                      offset:int=0, return_df:bool=True) -> Union[pd.DataFrame, Tuple[ListRecords, dict]]:
        df_row = self.datastore_search_sql(sql, limit=1, search_all=False, offset=offset, params=params, return_df=return_df)
        return df_row

    def datastore_search_sql_fields_type_dict(self, sql:str, *, params:dict=None) -> OrderedDict:
        document, fields_dict = self.datastore_search_sql_find_one(sql, offset=0, params=params, return_df=False)
        return fields_dict

    def datastore_search_sql_row_count(self, sql:str, *, params:dict=None) -> int:
        df_row = self.datastore_search_sql_find_one(sql, offset=0, params=params, return_df=True)
        return df_row.attrs["total"]

    def datastore_search_find_one(self, resource_id:str, *, filters:dict=None, q:str=None, distinct:bool=None,
                                  fields:List[str]=None, offset:int=0, return_df:bool=True) \
            -> Union[pd.DataFrame, ListRecords, Any, List[CkanActionResponse]]:
        """
        Request one result for a query

        :param resource_id: resource id
        :return:
        """
        # resource_info = self.get_resource_info_or_request(resource_id)
        # return resource_info.datastore_info.row_count
        df_row = self.datastore_search(resource_id, limit=1, search_all=False, filters=filters, q=q, distinct=distinct,
                                       fields=fields, offset=offset, return_df=return_df)
        return df_row

    def datastore_search_fields_type_dict(self, resource_id:str, *,
                                          filters:dict=None, q:str=None, distinct:bool=None, fields:List[str]=None,
                                          request_missing:bool=True, error_not_mapped:bool=False,
                                          error_not_found:bool=True) -> OrderedDict:
        if fields is None:
            # if no field restriction was provided, refer to the fields of the DataStore
            fields_list = self.get_datastore_fields_or_request(resource_id, return_list=True,
                                                               request_missing=request_missing,
                                                               error_not_mapped=error_not_mapped,
                                                               error_not_found=error_not_found)
            return CkanApiReadOnly.read_fields_type_dict(fields_list)
        else:
            document, fields_dict = self.datastore_search_find_one(resource_id, filters=filters, q=q, distinct=distinct,
                                                                   fields=fields, return_df=False)
            return fields_dict

    def datastore_search_row_count(self, resource_id:str, *, filters:dict=None, q:str=None, distinct:bool=None,
                                   fields:List[str]=None) -> int:
        """
        Request the number of rows in a DataStore

        :param resource_id: resource id
        :return:
        """
        df_row = self.datastore_search_find_one(resource_id, filters=filters, q=q, distinct=distinct,
                                                fields=fields, return_df=True)
        return df_row.attrs["total"]

    def test_sql_capabilities(self, *, raise_error:bool=False) -> bool:
        """
        Test the availability of the API datastore_search_sql

        :return:
        """
        try:
            self.api_help_show("datastore_search_sql", print_output=False)
            return True
        except CkanNotFoundError:
            if raise_error:
                raise CkanSqlCapabilityError(self, CkanActionResponse(requests.Response()))
            return False


    ## Resource download by direct link (FileStore) -----------------------------------------------
    def resource_download(self, resource_id:str, *, method:str=None,
                          proxies:dict=None, headers:dict=None, auth: Union[AuthBase, Tuple[str,str]]=None, verify:Union[bool,str,None]=None) \
            -> Tuple[CkanResourceInfo, Union[requests.Response,None]]:
        """
        Uses the link provided in resource_show to download a resource.

        :param resource_id: resource id
        :return:
        """
        resource_info = self.get_resource_info_or_request(resource_id)
        url = resource_info.download_url
        if len(url) == 0:
            return resource_info, None
        response = self.download_url_proxy(url, method=method, auth_if_ckan=ckan_request_proxy_default_auth_if_ckan,
                                           proxies=proxies, headers=headers, auth=auth, verify=verify)
        return resource_info, response

    def resource_download_test_head(self, resource_id:str, *, raise_error:bool=False,
                                    proxies:dict=None, headers:dict=None, auth: Union[AuthBase, Tuple[str,str]]=None, verify:Union[bool,str,None]=None) \
            -> Union[None,ContextErrorLevelMessage]:
        """
        This sends a HEAD request to the resource download url using the CKAN connexion parameters via resource_download.
        The resource is not downloaded but the headers indicate if the url is valid.

        :return: None if successful
        """
        resource_info = self.get_resource_info_or_request_of_id(resource_id)
        try:
            _, response = self.resource_download(resource_id, method="HEAD", proxies=proxies, headers=headers, auth=auth, verify=verify)
        except Exception as e:
            if raise_error:
                raise e from e
            return ContextErrorLevelMessage(f"Resource from URL {resource_info.name}", ErrorLevel.Error, f"Failed to query download url for resource id {resource_id}: {str(e)}")
        if response.ok and response.status_code == 200:
            return None
        else:
            if raise_error:
                raise RequestError(f"Failed to query download url for resource id {resource_id}: status {response.status_code} {response.reason}")
            return ContextErrorLevelMessage(f"Resource from URL {resource_info.name}", ErrorLevel.Error, f"Failed to query download url for resource id {resource_id}: status {response.status_code} {response.reason}")

    def resource_download_df(self, resource_id:str, *, method:str=None,
                          proxies:dict=None, headers:dict=None, auth: Union[AuthBase, Tuple[str,str]]=None, verify:Union[bool,str,None]=None) \
            -> Tuple[CkanResourceInfo, Union[pd.DataFrame,None]]:
        """
        Uses the link provided in resource_show to download a resource and interprets it as a DataFrame.

        :param resource_id: resource id
        :return:
        """
        resource_info, response = self.resource_download(resource_id, method=method, proxies=proxies, headers=headers, auth=auth, verify=verify)
        if response is None:
            return resource_info, None
        buffer = io.StringIO(response.content.decode())
        df = pd.read_csv(buffer, **df_download_read_csv_kwargs)
        self._rx_records_df_clean(df)
        return resource_info, df

    def map_file_resource_sizes(self, cancel_if_present:bool=True) -> None:
        for resource_id, resource_info in self.map.resources.items():
            if resource_info.download_url:
                if not (cancel_if_present and resource_info.download_size_mb is not None):
                    _, response = self.resource_download(resource_id, method="HEAD")
                    content_length = int(response.headers.get("content-length", None))  # raise error if not found or bad format
                    resource_info.download_size_mb = bytes_to_megabytes(content_length)


    ## Mapping of resource aliases from table
    def list_datastore_aliases(self) -> List[CkanAliasInfo]:
        alias_resource_id = "_table_metadata"  # resource name of table containing CKAN aliases
        alias_list_dict, _ = self.datastore_search(alias_resource_id, search_all=True, return_df=False, format="objects", search_method=True)
        alias_list = [CkanAliasInfo(alias_dict) for alias_dict in alias_list_dict]
        for alias_info in alias_list:
            if alias_info.alias_of is not None:
                self.map.resource_alias_index[alias_info.name] = alias_info.alias_of
        return alias_list

    def map_resources(self, package_list:Union[str, List[str]]=None, *, params:dict=None,
                      datastore_info:bool=None, resource_view_list:bool=None, organization_info:bool=None, license_list:bool=None,
                      only_missing:bool=True, error_not_found:bool=True,
                      owner_org:str=None) -> CkanMap:
        # overload including a call to list all aliases
        if len(self.map.resource_alias_index) == 0 and self.params.map_all_aliases:
            self.list_datastore_aliases()
        map = super().map_resources(package_list=package_list, params=params, datastore_info=datastore_info,
                              resource_view_list=resource_view_list, organization_info=organization_info,
                              license_list=license_list, only_missing=only_missing, error_not_found=error_not_found,
                              owner_org=owner_org)
        return map




#!python3
# -*- coding: utf-8 -*-
"""
Section of the package dedicated to the initialization of a CKAN package
"""

BUILDER_FILE_FORMAT_VERSION = "0.0.0"  # version of the Excel & JSON file format

from . import builder_aux
from . import builder_errors
from . import builder_field
from . import builder_resource
from . import builder_resource_multi_abc
from . import builder_resource_multi_file
from . import mapper_datastore
from . import builder_resource_datastore
from . import builder_resource_multi_datastore
from . import builder_resource_datastore_url
from . import builder_resource_datastore_unmanaged
from . import mapper_datastore_multi
from . import builder_resource_datastore_multi_abc
from . import builder_resource_datastore_multi_folder
from . import builder_resource_datastore_multi_harvester
from . import builder_resource_init
from . import builder_ckan
from . import builder_package_1_basic
from . import builder_package_2_harvesters
from . import builder_package_3_multi_threaded
from . import builder_package

from . import specific
from . import example

# usage shortcuts
from .builder_package import BuilderPackage
from .mapper_datastore_multi import RequestFileMapperIndexKeys
from .builder_resource_datastore_multi_abc import BuilderDataStoreMultiABC
from .builder_resource_datastore_multi_folder import BuilderDataStoreFolder





#!python3
# -*- coding: utf-8 -*-
"""
Package with helper function for CKAN requests using pandas DataFrames.
"""

# builder_file_format_version = "0.0.1"
try:
    from importlib.metadata import version, PackageNotFoundError
except ImportError:  # Python <3.8
    from importlib_metadata import version, PackageNotFoundError

try:
    __version__ = version("ckanapi_harvesters")
except PackageNotFoundError:
    __version__ = None


import os
self_dir = os.path.abspath(os.path.dirname(os.path.realpath(__file__)))


from . import auxiliary
from . import policies
from . import harvesters
from . import ckan_api
from . import builder
from . import reports

# usage shortcuts
from .auxiliary import CkanMap
from .policies import CkanPackageDataFormatPolicy
from .ckan_api import CkanApi, CKAN_API_VERSION
from .builder import BUILDER_FILE_FORMAT_VERSION
from .builder import BuilderPackage, BuilderDataStoreMultiABC, BuilderDataStoreFolder, RequestFileMapperIndexKeys





#!python3
# -*- coding: utf-8 -*-
"""
Tests to perform after the example package was uploaded
"""
from typing import Tuple
import os
import re
import getpass
import json

import pandas as pd
import numpy as np

from ckanapi_harvesters.auxiliary import CkanMap
from ckanapi_harvesters.builder.builder_package import BuilderPackage
from ckanapi_harvesters.ckan_api import CkanApi

from ckanapi_harvesters.builder.example import example_package_xls
self_dir = os.path.abspath(os.path.dirname(os.path.realpath(__file__)))


def run(ckan:CkanApi = None):
    BuilderPackage.unlock_external_code_execution()  # comment to test if the safety feature is enabled

    mdl = BuilderPackage.from_excel(example_package_xls)
    ckan = mdl.init_ckan(ckan)
    ckan.initialize_from_cli_args()
    ckan.input_missing_info(input_args_if_necessary=True, input_owner_org=True)
    ckan.set_verbosity(True)

    # Test re-encoding the Excel file from the loaded model
    example_package_xls_reencoded = os.path.abspath("builder_package_example-reencoded.xlsx")
    mdl.to_excel(example_package_xls_reencoded)

    # map package
    ckan.map_resources(mdl.package_name, datastore_info=True, license_list=True)
    map_init = ckan.map.copy()
    ckan.purge(purge_map=True)

    # Test using the model to update CKAN map (update_ckan_map)
    mdl.info_request_full(ckan)
    map_queried = ckan.map.copy()
    ckan.purge(purge_map=True)
    mdl.update_ckan_map(ckan)
    map_from_mdl = ckan.map.copy()

    # Test saving the map to a dictionary
    dict_map_queried = map_queried.to_dict()
    map_queried_from_dict = CkanMap.from_dict(dict_map_queried)

    dict_map_from_mdl = map_from_mdl.to_dict()
    map_from_mdl_from_dict = CkanMap.from_dict(dict_map_from_mdl)

    # test the function that recreates the Excel file from the online information
    ckan.purge(purge_map=True)
    mdl_api = BuilderPackage.from_ckan(ckan, mdl.package_name)
    example_package_xls_from_api = os.path.abspath("builder_package_example-from-api.xlsx")
    mdl_api.to_excel(example_package_xls_from_api)

    # export mdl to dict and re-import
    base_dir = os.path.abspath(".")
    mdl_dict = mdl.to_dict(base_dir=base_dir)
    mdl_from_dict = BuilderPackage.from_dict(mdl_dict, base_dir=base_dir)
    mdl_dict_bis = mdl_from_dict.to_dict(base_dir=base_dir)
    assert(mdl_dict == mdl_dict_bis)

    # test json serialization
    example_package_json_reencoded = os.path.abspath("builder_package_example-reencoded.json")
    with open(example_package_json_reencoded, "w") as json_file:
        json.dump(mdl_dict, json_file, indent=4)
        json_file.close()

    # test copy constructors
    ckan_copy = ckan.copy()
    mdl_copy = mdl.copy()
    mdl_copy_dict = mdl_copy.to_dict(base_dir=base_dir)
    assert(mdl_dict == mdl_copy_dict)

    print("Tests done.")


if __name__ == '__main__':
    ckan = CkanApi(None)
    ckan.initialize_from_cli_args()
    run(ckan)




#!python3
# -*- coding: utf-8 -*-
"""
Auxiliary functions
"""
from typing import List, Union
import os

def positive_end_index(end_index:Union[int,None], total:int) -> int:
    """
    Return stop index for a loop, following pythonic definition for slices (last index treated = end_index-1).
    If end_index is negative, the index is taken from the end of the slice. end_index = -1 means end just before the last element.
    """
    if end_index is None:
        return total
    elif end_index < 0:
        return max(0, total + end_index)
    else:
        return end_index




#!python3
# -*- coding: utf-8 -*-
"""
Harvester initialization from the options_string arguments
"""
from ckanapi_harvesters.harvesters.harvester_abc import TableHarvesterABC, DatasetHarvesterABC
from ckanapi_harvesters.harvesters.harvester_params import TableParams
from ckanapi_harvesters.harvesters.harvester_params import DatasetParams
from ckanapi_harvesters.harvesters.postgre_harvester import TableHarvesterPostgre, DatasetHarvesterPostgre
from ckanapi_harvesters.harvesters.pymongo_harvester import TableHarvesterMongoCollection, DatasetHarvesterMongoDatabase


def init_table_harvester_from_options_string(options_string:str, *, file_url_attr:str, base_dir:str=None) -> TableHarvesterABC:
    harvest_method = TableParams.parse_harvest_method(options_string)
    if harvest_method == "pymongo":
        return TableHarvesterMongoCollection.init_from_options_string(options_string, file_url_attr=file_url_attr, base_dir=base_dir)
    elif harvest_method == "postgre":
        return TableHarvesterPostgre.init_from_options_string(options_string, file_url_attr=file_url_attr, base_dir=base_dir)
    else:
        raise NotImplementedError(f"harvester method {harvest_method} not implemented")


def init_dataset_harvester_from_options_string(options_string:str, *, base_dir:str=None) -> DatasetHarvesterABC:
    harvest_method = DatasetParams.parse_harvest_method(options_string)
    if harvest_method == "pymongo":
        return DatasetHarvesterMongoDatabase.init_from_options_string(options_string, base_dir=base_dir)
    elif harvest_method == "postgre":
        return DatasetHarvesterPostgre.init_from_options_string(options_string, base_dir=base_dir)
    else:
        raise NotImplementedError(f"harvester method {harvest_method} not implemented")



#!python3
# -*- coding: utf-8 -*-
"""
Tests to perform after the example package was uploaded
"""
from typing import Tuple
import os
import re
import getpass
import json

import pandas as pd
import numpy as np

from ckanapi_harvesters.auxiliary import CkanMap
from ckanapi_harvesters.builder.builder_package import BuilderPackage
from ckanapi_harvesters.ckan_api import CkanApi

from ckanapi_harvesters.builder.example import example_package_xls
self_dir = os.path.abspath(os.path.dirname(os.path.realpath(__file__)))


if __name__ == '__main__':
    BuilderPackage.unlock_external_code_execution()  # comment to test if the safety feature is enabled

    mdl = BuilderPackage.from_excel(example_package_xls)

    ckan = CkanApi()
    ckan = mdl.init_ckan(ckan)
    # mdl.update_ckan_map(ckan)
    # map_from_mdl = ckan.map.copy()

    # Test re-encoding the Excel file from the loaded model
    example_package_xls_reencoded = os.path.abspath("builder_package_example-reencoded.xlsx")
    mdl.to_excel(example_package_xls_reencoded)

    # export mdl to dict and re-import
    base_dir = os.path.abspath(".")
    mdl_dict = mdl.to_dict(base_dir=base_dir)
    mdl_from_dict = BuilderPackage.from_dict(mdl_dict, base_dir=base_dir)
    mdl_dict_bis = mdl_from_dict.to_dict(base_dir=base_dir)
    assert(mdl_dict == mdl_dict_bis)

    # test json serialization
    example_package_json_reencoded = os.path.abspath("builder_package_example-reencoded.json")
    with open(example_package_json_reencoded, "w") as json_file:
        json.dump(mdl_dict, json_file, indent=4)

    # test copy constructors
    ckan_copy = ckan.copy()
    mdl_copy = mdl.copy()
    mdl_copy_dict = mdl_copy.to_dict(base_dir=base_dir)
    assert(mdl_dict == mdl_copy_dict)

    print("Tests done.")





#!python3
# -*- coding: utf-8 -*-
"""
File format base class
"""
from abc import ABC, abstractmethod
from typing import Union, Dict
import io

import pandas as pd

from ckanapi_harvesters.auxiliary.ckan_model import CkanField
from ckanapi_harvesters.auxiliary.list_records import ListRecords


class FileFormatABC(ABC):
    @abstractmethod
    def read_file(self, file_path: str, fields: Union[Dict[str, CkanField],None]) -> Union[pd.DataFrame, ListRecords]:
        raise NotImplementedError()

    @abstractmethod
    def read_buffer(self, buffer: io.IOBase, fields: Union[Dict[str, CkanField],None]) -> Union[pd.DataFrame, ListRecords]:
        raise NotImplementedError()

    @abstractmethod
    def write_file(self, df: Union[pd.DataFrame, ListRecords], file_path: str, fields: Union[Dict[str, CkanField],None]) -> None:
        raise NotImplementedError()

    @abstractmethod
    def write_in_memory(self, df: Union[pd.DataFrame, ListRecords], fields: Union[Dict[str, CkanField],None]) -> bytes:
        raise NotImplementedError()

    @abstractmethod
    def copy(self):
        raise NotImplementedError()

    def __copy__(self):
        return self.copy()




#!python3
# -*- coding: utf-8 -*-
"""
Harvest from a mongo database using pymongo
"""
from typing import Union, List, Any, Dict
from types import SimpleNamespace
from collections import OrderedDict
import json
import argparse


try:
    import pymongo
    import pymongo.client_session
    import pymongo.database
except ImportError:
    pymongo = SimpleNamespace(MongoClient=None, client_session=SimpleNamespace(ClientSession=None),
                              database=SimpleNamespace(Database=None), collection=SimpleNamespace(Collection=None))


from ckanapi_harvesters.harvesters.harvester_errors import (HarvesterRequirementError, HarvesterArgumentRequiredError, ResourceNotFoundError)
from ckanapi_harvesters.harvesters.harvester_abc import TableHarvesterABC, DatasetHarvesterABC, DatabaseHarvesterABC
from ckanapi_harvesters.harvesters.harvester_model import TableMetadata, DatasetMetadata
from ckanapi_harvesters.harvesters.harvester_params import DatasetParams, DatabaseParams
from ckanapi_harvesters.harvesters.pymongo_data_cleaner import pymongo_default_data_cleaner, pymongo_default_df_conversion
from ckanapi_harvesters.harvesters.pymongo_data_cleaner import mongodb_keep_id_column_trace, mongodb_id_new_column
from ckanapi_harvesters.harvesters.pymongo_params import TableParamsMongoCollection
from ckanapi_harvesters.auxiliary.urls import url_join
from ckanapi_harvesters.auxiliary.ckan_auxiliary import ssl_arguments_decompose, assert_or_raise
from ckanapi_harvesters.auxiliary.ckan_errors import UrlError
from ckanapi_harvesters.harvesters.data_cleaner.data_cleaner_abc import CkanDataCleanerABC
from ckanapi_harvesters.auxiliary.error_level_message import ContextErrorLevelMessage, ErrorLevel


mongodb_excluded_collections = {"system.profile"}


class DatabaseHarvesterMongoServer(DatabaseHarvesterABC):
    """
    This class manages the connection to a MongoDB server.
    It can list datasets (MongoDB databases) but this call could lead to an error.
    """
    def __init__(self, params:DatabaseParams=None):
        super().__init__(params)
        if pymongo.MongoClient is None:
            raise HarvesterRequirementError("pymongo", "pymongo")
        self.params.harvest_method = "Pymongo"
        self.mongo_client: Union[pymongo.MongoClient,None] = None
        self.mongo_session: Union[pymongo.client_session.ClientSession,None] = None
        if self.params.auth_url is None and self.params.port is None and self.params.host is None:
            raise HarvesterArgumentRequiredError("auth-url", "pymongo", "This argument defines the url used to authenticate.")

    @staticmethod
    def init_from_options_string(options_string:str, base_dir:str=None) -> "DatabaseHarvesterMongoServer":
        params = DatabaseParams()
        params.parse_options_string(options_string, base_dir=base_dir)
        return DatabaseHarvesterMongoServer(params)

    def copy(self, *, dest=None):
        if dest is None:
            dest = DatabaseHarvesterMongoServer()
        return super().copy(dest=dest)

    def connect(self, *, cancel_if_connected:bool=True) -> Any:
        if cancel_if_connected and self.mongo_client is not None:
            return self.mongo_client
        else:
            if self.mongo_client is not None:
                self.mongo_session.end_session()
                self.mongo_client.close()
                self.mongo_session = None
                self.mongo_client = None
            ssl, ssl_certfile = ssl_arguments_decompose(self.params.verify_ca)
            auth_url = self.params.auth_url
            if auth_url is None:
                if self.params.url is not None:
                    auth_url = self.params.url
                elif self.params.host is not None:
                    auth_url = f"mongodb://{self.params.host}"
                    if self.params.port is not None:
                        auth_url += f":{self.params.port}"
                else:
                    raise UrlError("No Mongo URL provided")
                if self.params.auth_url_suffix is not None:
                    auth_url = url_join(auth_url, self.params.auth_url_suffix)
                self.params.auth_url = auth_url
            self.mongo_client = pymongo.MongoClient(auth_url, username=self.params.login.username, password=self.params.login.password,
                                                    ssl=ssl, tlscafile=ssl_certfile,
                                                    timeoutMS=self.params.timeout*1000.0 if self.params.timeout is not None else None)
            self.mongo_session = self.mongo_client.start_session()
            if self.params.host is None and self.params.port is None:
                # complete with host and port parsed by MongoClient
                mongo_address = self.mongo_client.address
                if mongo_address is not None:
                    self.params.host, self.params.port = mongo_address
            return self.mongo_client

    def is_connected(self) -> bool:
        return self.mongo_client is not None

    def disconnect(self) -> None:
        if self.mongo_client is not None:
            if self.mongo_session is not None:
                self.mongo_session.end_session()
            self.mongo_client.close()
            self.mongo_client = None
            self.mongo_session = None

    def check_connection(self, *, new_connection:bool=False, raise_error:bool=False) -> Union[None, ContextErrorLevelMessage]:
        try:
            self.connect(cancel_if_connected=not new_connection)
            # the following line requires specific admin rights (unexpected)
            # remote_collections = self.mongo_client.list_database_names(self.mongo_session)  # this tests the database connection
        except Exception as e:
            if raise_error:
                raise e from e
            else:
                return ContextErrorLevelMessage("Mongo Harvester", ErrorLevel.Error, f"Failed to connect to {self.params.auth_url}: {e}")

    def get_dataset_harvester(self, dataset_name: str) -> "DatasetHarvesterMongoDatabase":
        params_dataset = self.params.copy(dest=DatasetParams())
        params_dataset.dataset = dataset_name
        dataset_harvester = DatasetHarvesterMongoDatabase(params_dataset)
        self.copy(dest=dataset_harvester)
        dataset_harvester.params = params_dataset
        dataset_harvester._finalize_connection()
        return dataset_harvester

    def list_datasets(self, return_metadata: bool = True) -> Union[List[str], OrderedDict[str, DatasetMetadata]]:
        # this would raise an unauthorized error !
        self.connect()
        dataset_list = self.mongo_client.list_database_names(self.mongo_session)
        if return_metadata:
            return OrderedDict([(name, self.get_dataset_harvester(name).query_dataset_metadata()) for name in dataset_list])
        else:
            return dataset_list


class DatasetHarvesterMongoDatabase(DatabaseHarvesterMongoServer, DatasetHarvesterABC):
    """
    A CKAN dataset corresponds to a MongoDB database (set of collections).
    """
    def __init__(self, params:DatasetParams=None):
        super().__init__(params)
        self.mongo_database: Union[pymongo.database.Database,None] = None
        self.dataset_metadata: Union[DatasetMetadata, None] = None  # DatasetHarvesterABC
        # use database attribute if dataset is not specified (ambiguity on name - database attribute is not used above)
        if self.params.dataset is None:
            self.params.dataset = self.params.database
        if self.params.dataset is None:
            raise HarvesterArgumentRequiredError("dataset", "pymongo", "This argument defines the mongo database used")

    @staticmethod
    def init_from_options_string(options_string:str, base_dir:str=None) -> "DatasetHarvesterMongoDatabase":
        params = DatasetParams()
        params.parse_options_string(options_string, base_dir=base_dir)
        return DatasetHarvesterMongoDatabase(params)

    def _finalize_connection(self):
        if super().is_connected() and self.mongo_database is None:
            # remote_datasets = self.mongo_client.list_database_names(self.mongo_session)
            # assert_or_raise(self.params.dataset in remote_datasets, ResourceNotFoundError("Database", self.params.dataset, self.params.auth_url))
            self.mongo_database = self.mongo_client[self.params.dataset]

    def connect(self, *, cancel_if_connected:bool=True) -> Any:
        if not (cancel_if_connected and self.is_connected()):
            super().connect(cancel_if_connected=cancel_if_connected)
            self._finalize_connection()
        return self.mongo_client

    def is_connected(self) -> bool:
        return super().is_connected()

    def disconnect(self) -> None:
        if super().is_connected():
            self.mongo_database = None
            super().disconnect()

    def check_connection(self, *, new_connection: bool = False, raise_error: bool = False) -> Union[None, ContextErrorLevelMessage]:
        try:
            super().check_connection(new_connection=new_connection, raise_error=raise_error)
            remote_collections = self.mongo_database.list_collection_names(self.mongo_session)  # this tests the database connection
        except Exception as e:
            if raise_error:
                raise e from e
            else:
                return ContextErrorLevelMessage("Mongo Harvester", ErrorLevel.Error,
                                                f"Failed to connect to {self.params.auth_url}: {e}")
        if self.mongo_database is None:
            return ContextErrorLevelMessage("Mongo Harvester", ErrorLevel.Error, f"Failed to connect to {self.params.auth_url}: <no error message>")
        else:
            return None

    def query_dataset_metadata(self, cancel_if_present:bool=True) -> DatasetMetadata:
        self.connect()
        if cancel_if_present and self.dataset_metadata is not None:
            return self.dataset_metadata
        else:
            self.dataset_metadata = DatasetMetadata()
            self.dataset_metadata.name = self.mongo_database.name
            self.dataset_metadata.tables = self.list_tables(return_metadata=True)
            return self.dataset_metadata

    def clean_dataset_metadata(self) -> DatasetMetadata:
        return self.query_dataset_metadata().copy()

    def get_table_harvester(self, table_name:str) -> "TableHarvesterMongoCollection":
        params_table = self.params.copy(dest=TableParamsMongoCollection())
        if self.params.options_string is not None:
            # reparse options_string for table-specific arguments
            params_table.parse_options_string(self.params.options_string, base_dir=self.params.base_dir)
        params_table.table = table_name
        table_harvester = TableHarvesterMongoCollection(params_table)
        self.copy(dest=table_harvester)
        table_harvester.params = params_table
        table_harvester._finalize_connection()
        return table_harvester

    def list_tables(self, return_metadata:bool=True) -> Union[List[str], OrderedDict[str, TableMetadata]]:
        self.connect()
        remote_collections = [collection_name for collection_name in self.mongo_database.list_collection_names(session=self.mongo_session) if collection_name not in mongodb_excluded_collections]
        if return_metadata:
            return OrderedDict([(name, self.get_table_harvester(name).query_table_metadata()) for name in remote_collections])
        else:
            return remote_collections


class TableHarvesterMongoCollection(DatasetHarvesterMongoDatabase, TableHarvesterABC):
    """
    A table (CKAN DataStore) corresponds to a MongoDB collection.
    """
    _default_upload_fun = pymongo_default_df_conversion
    _default_primary_key = [mongodb_id_new_column]

    def __init__(self, params:TableParamsMongoCollection=None):
        super().__init__(params)
        self.params: TableParamsMongoCollection = params
        self.mongo_collection: Union[pymongo.collection.Collection,None] = None
        self.table_metadata: Union[TableMetadata, None] = None  # TableHarvesterABC
        if self.params.file_url_attr is not None:
            # File/URL attribute has priority over CLI
            self.params.table = self.params.file_url_attr
        if self.params.table is None:
            raise HarvesterArgumentRequiredError("table", "pymongo", "This argument defines the mongo collection used")

    @staticmethod
    def init_from_options_string(options_string:str, *, base_dir:str=None, file_url_attr:str=None) -> "TableHarvesterMongoCollection":
        params = TableParamsMongoCollection()
        params.parse_options_string(options_string, file_url_attr=file_url_attr, base_dir=base_dir)
        return TableHarvesterMongoCollection(params)

    def copy(self, *, dest=None):
        if dest is None:
            dest = TableHarvesterMongoCollection()
        super().copy(dest=dest)
        return dest

    def disconnect(self) -> None:
        if super().is_connected():
            self.mongo_collection = None
            super().disconnect()

    def _finalize_connection(self):
        super()._finalize_connection()
        if super().is_connected() and self.mongo_collection is None:
            mongo_database = self.mongo_database
            remote_collections = mongo_database.list_collection_names()
            collection_name = self.params.table
            assert_or_raise(collection_name in remote_collections, ResourceNotFoundError("Collection", self.params.dataset + '.' + collection_name, self.params.auth_url))
            collection = mongo_database[collection_name]
            self.mongo_collection = collection

    def connect(self, *, cancel_if_connected:bool=True) -> Any:
        if not (cancel_if_connected and self.is_connected()):
            super().connect()
            self._finalize_connection()
        return self.mongo_client

    def check_connection(self, *, new_connection:bool=False, raise_error:bool=False) -> Union[None, ContextErrorLevelMessage]:
        super().check_connection(new_connection=new_connection, raise_error=raise_error)
        if self.mongo_collection is None:
            return ContextErrorLevelMessage("Mongo Harvester", ErrorLevel.Error, f"Failed to connect to {self.params.auth_url}: <no error message>")
        else:
            return None

    def query_table_metadata(self, cancel_if_present:bool=True) -> TableMetadata:
        self.connect()
        if cancel_if_present and self.table_metadata is not None:
            return self.table_metadata
        else:
            # TODO: query at least primary key and indexes
            self.table_metadata = TableMetadata()
            index_dict = self.mongo_collection.index_information(session=self.mongo_session)
            self.table_metadata.name = self.mongo_collection.name
            self.table_metadata.indexes = sum([[key[0] for key in index["key"]] for index in index_dict.values()], [])
            return self.table_metadata

    def clean_table_metadata(self) -> TableMetadata:
        clean_metadata = self.query_table_metadata().copy()
        if clean_metadata.indexes is not None:
            i_rm = []
            for i, name in enumerate(clean_metadata.indexes):
                if name == "_id":
                    if mongodb_keep_id_column_trace:
                        clean_metadata.indexes[i] = mongodb_id_new_column
                    else:
                        i_rm.append(i)
                elif "." in name:
                    i_rm.append(i)
            for i in reversed(i_rm):
                clean_metadata.indexes.pop(i)
        clean_metadata.indexes = None  # finally, do not specify indexes at all
        return clean_metadata

    def get_default_primary_key(self) -> List[str]:
        table_metadata = self.query_table_metadata()
        if table_metadata.primary_key is not None:
            return table_metadata.primary_key
        else:
            return TableHarvesterMongoCollection._default_primary_key

    def get_default_data_cleaner(self) -> Union[CkanDataCleanerABC, None]:
        data_cleaner = pymongo_default_data_cleaner()
        data_cleaner.param_mongodb_dbref_as_one_column = not self.params.dbref_expand
        return data_cleaner

    def list_queries(self, *, new_connection:bool=False) -> List[Dict[str,Any]]:
        self.connect(cancel_if_connected=not new_connection)
        assert(self.mongo_collection is not None)
        query = json.loads(self.params.query_string) if self.params.query_string is not None else {}
        if self.params.verbose_harvester:
            print(f"Counting documents of table {self.params.table}")
        # num_rows = self.mongo_collection.count_documents(query, session=self.mongo_session)
        num_rows = self.mongo_collection.estimated_document_count()
        num_queries = num_rows // self.params.limit + 1
        if self.params.single_request:
            return [OrderedDict([("$match", query), ("$skip", i * self.params.limit), ("$limit", self.params.limit)]) for i in range(1)]
        else:
            queries_exact = [OrderedDict([("$match", query), ("$skip", i * self.params.limit), ("$limit", self.params.limit)]) for i in range(num_queries)]
            query_extra = OrderedDict([("$match", query), ("$skip", num_queries * self.params.limit)])
            return queries_exact + [query_extra]

    def query_data(self, query:Dict[str,Any]) -> List[dict]:
        assert(self.mongo_collection is not None)
        if isinstance(query, str):
            query = json.loads(query)
        if self.params.verbose_harvester:
            print(f"Pymongo request {query} on table {self.params.table}")
        cursor = self.mongo_collection.find(query["$match"], session=self.mongo_session).skip(query["$skip"])
        if "$limit" in query.keys():
            cursor = cursor.limit(query["$limit"])
        documents = list(cursor)
        return documents




#!python3
# -*- coding: utf-8 -*-
"""
Code to upload metadata to the CKAN server to create/update an existing package
The metadata is defined by the user in an Excel worksheet
This file implements the basic resources. See builder_datastore for specific functions to initiate datastores.
"""
from concurrent.futures import ThreadPoolExecutor
import threading
from threading import current_thread
from typing import Any, Generator, Union, Callable, Set, List, Dict, Tuple
from abc import ABC, abstractmethod
import io
import os
import glob
import fnmatch
from warnings import warn
import copy

import pandas as pd
import requests

from ckanapi_harvesters.auxiliary.error_level_message import ContextErrorLevelMessage, ErrorLevel
from ckanapi_harvesters.auxiliary.ckan_auxiliary import _string_from_element
from ckanapi_harvesters.auxiliary.ckan_defs import ckan_tags_sep
from ckanapi_harvesters.ckan_api import CkanApi
from ckanapi_harvesters.auxiliary.ckan_model import UpsertChoice, CkanResourceInfo
from ckanapi_harvesters.builder.builder_aux import positive_end_index
from ckanapi_harvesters.builder.builder_errors import ResourceFileNotExistMessage
from ckanapi_harvesters.builder.builder_field import BuilderField
from ckanapi_harvesters.builder.builder_resource import BuilderResourceABC

multi_file_exclude_other_files:bool = True


def default_progress_callback(index:int, total:int, info:Any, *, context:str=None, **kwargs) -> None:
    if context is None:
        context = ""
    if index == total:
        # info is None
        print(f"{context} Finished {index}/{total} (100%)")
    elif info is None:
        print(f"{context} Request {index}/{total} ({index/total*100.0:.2f}%)")
    else:
        if isinstance(info, str):
            info_str = info
        elif isinstance(info, pd.DataFrame):
            if "source" in info.attrs.keys():
                info_str = str(info.attrs["source"])
            else:
                info_str = "<DataFrame>"
        elif isinstance(info, list):
            info_str = "<records>"
        else:
            info_str = str(info)
        print(f"{context} Request {index}/{total} ({index/total*100.0:.2f}%): " + info_str)


class BuilderMultiABC(ABC):
    def __init__(self):
        self.progress_callback: Union[Callable[[int, int, Any], None], None] = default_progress_callback
        self.progress_callback_kwargs: dict = {}
        self.stop_event = threading.Event()
        self.thread_ckan: Dict[str, CkanApi] = {}
        self.enable_multi_threaded_upload:bool = True
        self.enable_multi_threaded_download:bool = True
        # from Resource (for code validation)
        self.name:str = ""
        self.enable_download:bool = True

    def copy(self, *, dest=None):
        dest.progress_callback = self.progress_callback
        dest.progress_callback_kwargs = copy.deepcopy(self.progress_callback_kwargs)
        dest.enable_multi_threaded_upload = self.enable_multi_threaded_upload
        dest.enable_multi_threaded_download = self.enable_multi_threaded_download
        # do not copy stop_event
        return dest

    def _call_progress_callback(self, index:int, total:int, *, info:Any=None, context:str=None) -> None:
        if self.progress_callback is not None:
            self.progress_callback(index, total, info=info, context=context, **self.progress_callback_kwargs)

    def _prepare_for_multithreading(self, ckan: CkanApi):
        self.stop_event.clear()
        self.thread_ckan.clear()

    def _init_thread(self, ckan: CkanApi):
        thread_name = current_thread().name
        ckan_thread = ckan.copy(new_identifier=thread_name)
        ckan_thread.prepare_for_multithreading(True)  # prepare CKAN object for multi-threading
        self.thread_ckan[thread_name] = ckan_thread

    def _terminate_thread(self):
        for ckan in self.thread_ckan.values():
            ckan.disconnect()
        self.thread_ckan.clear()


    ## upload -----------------------------------------------------------------
    @abstractmethod
    def init_local_files_list(self, resources_base_dir:str, cancel_if_present:bool=True, **kwargs) -> List[str]:
        """
        Behavior to list parts of an upload.
        """
        raise NotImplementedError()

    @abstractmethod
    def get_local_file_len(self) -> int:
        """
        Get the number of parts of the upload.
        """
        raise NotImplementedError()

    @abstractmethod
    def get_local_file_generator(self, resources_base_dir:str, **kwargs) -> Generator[Any, None, None]:
        """
        Returns an iterator over the parts of the upload.
        """
        raise NotImplementedError()

    @abstractmethod
    def upload_request_final(self, ckan:CkanApi, *, force:bool=False) -> None:
        raise NotImplementedError()

    @abstractmethod
    def _unit_upload_apply(self, *, ckan:CkanApi, file:Any,
                           index:int, start_index:int, end_index:int, total:int, **kwargs) -> Any:
        """
        Unitary function deciding whether to perform upload and making the steps for the upload.
        """
        raise NotImplementedError()

    def upload_request_full(self, ckan:CkanApi, resources_base_dir:str, *,
                            threads:int=1, external_stop_event=None,
                            start_index:int=0, end_index:int=None, **kwargs) -> None:
        """
        Perform all the upload requests.

        :param ckan:
        :param resources_base_dir:
        :param threads:
        :param external_stop_event:
        :param only_missing:
        :param start_index:
        :param end_index:
        :return:
        """
        if threads < 0:
            # cancel large uploads in this case
            return None
        elif (threads is None or threads > 1) and self.enable_multi_threaded_upload:
            return self.upload_request_full_multi_threaded(ckan=ckan, resources_base_dir=resources_base_dir,
                                                           threads=threads, external_stop_event=external_stop_event,
                                                           start_index=start_index, end_index=end_index, **kwargs)
        else:
            self.init_local_files_list(resources_base_dir=resources_base_dir, cancel_if_present=True, **kwargs)
            if ckan.params.verbose_extra:
                print(f"Launching single-threaded upload of multi-file resource {self.name}")
            total = self.get_local_file_len()
            end_index = positive_end_index(end_index, total)
            for index, file_path in enumerate(self.get_local_file_generator(resources_base_dir=resources_base_dir, **kwargs)):
                if external_stop_event is not None and external_stop_event.is_set():
                    print(f"{ckan.identifier} Interrupted")
                    return
                self._unit_upload_apply(ckan=ckan, file=file_path, index=index,
                                        start_index=start_index, end_index=end_index, total=total, **kwargs)
            self._call_progress_callback(total, total, context=f"{ckan.identifier} single-thread upload")
            # at last, apply final actions:
            self.upload_request_final(ckan)

    def upload_request_graceful(self, ckan:CkanApi, file_path: str, *, index:int,
                                external_stop_event=None,
                                start_index:int=0, end_index:int=None, **kwargs) -> None:
        """
        Calls upload_file with checks specific to multi-threading.

        :return:
        """
        # ckan.session_reset()
        # ckan.identifier = current_thread().name
        ckan = self.thread_ckan[current_thread().name]
        total = self.get_local_file_len()
        end_index = positive_end_index(end_index, total)
        if self.stop_event.is_set():
            return
        if external_stop_event is not None and external_stop_event.is_set():
            print(f"{ckan.identifier} Interrupted")
            return
        try:
            self._unit_upload_apply(ckan=ckan, file=file_path, index=index,
                                    start_index=start_index, end_index=end_index, total=total, **kwargs)
        except Exception as e:
            self.stop_event.set()  # Ensure all threads stop
            if ckan.params.verbose_extra:
                print(f"Stopping all threads because an exception occurred in thread: {e}")
            raise e from e

    def upload_request_full_multi_threaded(self, ckan:CkanApi, resources_base_dir:str,
                            threads:int=1, external_stop_event=None,
                            start_index:int=0, end_index:int=None, **kwargs):
        """
        Multi-threaded implementation of upload_request_full, using ThreadPoolExecutor.
        """
        self.init_local_files_list(resources_base_dir=resources_base_dir, cancel_if_present=True, **kwargs)
        self._prepare_for_multithreading(ckan)
        try:
            with ThreadPoolExecutor(max_workers=threads, initializer=self._init_thread, initargs=(ckan,)) as executor:
                if ckan.params.verbose_extra:
                    print(f"Launching multi-threaded upload of multi-file resource {self.name}")
                futures = [executor.submit(self.upload_request_graceful, ckan=ckan, file_path=file_path, index=index,
                                           start_index=start_index, end_index=end_index, external_stop_event=external_stop_event,
                                           **kwargs)
                           for index, file_path in enumerate(self.get_local_file_generator(resources_base_dir=resources_base_dir, **kwargs))]
                for future in futures:
                    future.result()  # This will propagate the exception
            total = self.get_local_file_len()
            self._call_progress_callback(total, total, context=f"{ckan.identifier} multi-thread upload")
        except Exception as e:
            self.stop_event.set()  # Ensure all threads stop
            if ckan.params.verbose_extra:
                print(f"Stopping all threads because an exception occurred: {e}")
            raise e from e
        finally:
            # self.stop_event.set()  # Ensure all threads stop
            if ckan.params.verbose_extra:
                print("End of multi-threaded upload...")
        # at last, apply final actions:
        self._terminate_thread()
        self.upload_request_final(ckan)

    ## download -------------------------------------------------------------
    @abstractmethod
    def init_download_file_query_list(self, ckan: CkanApi, out_dir: str, cancel_if_present: bool = True, **kwargs) -> List[Any]:
        """
        Determine the list of queries to download to reconstruct the uploaded parts.
        By default, the unique combinations of the first columns of the primary key are used.
        """
        raise NotImplementedError()

    @abstractmethod
    def get_file_query_generator(self) -> Generator[Any, Any, None]:
        """
        Returns an iterator on all the file_queries.
        """
        raise NotImplementedError()

    @abstractmethod
    def get_file_query_len(self) -> int:
        """
        Returns the total number of file_queries.
        """
        raise NotImplementedError()

    @abstractmethod
    def download_file_query_item(self, ckan: CkanApi, out_dir: str, file_query_item: Any) -> Any:
        """
        Download the file_query item with the its arguments
        """
        raise NotImplementedError()

    @abstractmethod
    def download_request_generator(self, ckan: CkanApi, out_dir: str) -> Generator[Any, Any, None]:
        """
        Generator to apply treatments after each request (single-threaded).

        :param ckan:
        :param out_dir:
        :return:
        """
        raise NotImplementedError()

    @abstractmethod
    def _unit_download_apply(self, ckan:CkanApi, file_query_item:Any, out_dir:str,
                           index:int, start_index:int, end_index:int, total:int, **kwargs) -> Any:
        """
        Unitary function deciding whether to perform download and making the steps for the request.
        """
        raise NotImplementedError()

    def download_request_full(self, ckan: CkanApi, out_dir: str, threads:int=1, external_stop_event=None,
                              start_index:int=0, end_index:int=None, force:bool=False, **kwargs) -> None:

        if (not self.enable_download) and (not force):
            msg = f"Did not download resource {self.name} because download was disabled."
            warn(msg)
            return None
        if threads < 0:
            # do not download large datasets in this case
            return None
        elif (threads is None or threads > 1) and self.enable_multi_threaded_download:
            return self.download_request_full_multi_threaded(ckan=ckan, out_dir=out_dir,
                                                             threads=threads, external_stop_event=external_stop_event,
                                                             start_index=start_index, end_index=end_index, **kwargs)
        else:
            self.init_download_file_query_list(ckan=ckan, out_dir=out_dir, cancel_if_present=True, **kwargs)
            if ckan.params.verbose_extra:
                print(f"Launching single-threaded download of multi-file resource {self.name}")
            total = self.get_file_query_len()
            end_index = positive_end_index(end_index, total)
            for index, file_query_item in enumerate(self.get_file_query_generator()):
                if external_stop_event is not None and external_stop_event.is_set():
                    print(f"{ckan.identifier} Interrupted")
                    return
                self._unit_download_apply(ckan=ckan, file_query_item=file_query_item, out_dir=out_dir,
                                          index=index, start_index=start_index, end_index=end_index, total=total, **kwargs)
            self._call_progress_callback(total, total, context=f"{ckan.identifier} single-thread download")

    def download_file_query_item_graceful(self, ckan: CkanApi, out_dir: str, file_query_item: Any, index:int,
                                          external_stop_event=None, start_index:int=0, end_index:int=None, **kwargs) -> None:
        """
        Implementation of download_file_query_item with checks for a multi-threaded download.
        """
        # ckan.session_reset()
        # ckan.identifier = current_thread().name
        ckan = self.thread_ckan[current_thread().name]
        total = self.get_file_query_len()
        end_index = positive_end_index(end_index, total)
        if self.stop_event.is_set():
            return
        if external_stop_event is not None and external_stop_event.is_set():
            print(f"{ckan.identifier} Interrupted")
            return
        try:
            self._unit_download_apply(ckan=ckan, file_query_item=file_query_item, out_dir=out_dir,
                                      index=index, start_index=start_index, end_index=end_index, total=total, **kwargs)
        except Exception as e:
            self.stop_event.set()  # Ensure all threads stop
            if ckan.params.verbose_extra:
                print(f"Stopping all threads because an exception occurred in thread: {e}")
            raise e from e

    def download_request_full_multi_threaded(self, ckan: CkanApi, out_dir: str,
                                             threads: int = None, external_stop_event=None,
                                             start_index:int=0, end_index:int=-1, **kwargs) -> None:
        """
        Multi-threaded implementation of download_request_full using ThreadPoolExecutor.
        """
        self.init_download_file_query_list(ckan=ckan, out_dir=out_dir, cancel_if_present=True, **kwargs)
        self._prepare_for_multithreading(ckan)
        try:
            with ThreadPoolExecutor(max_workers=threads, initializer=self._init_thread, initargs=(ckan,)) as executor:
                if ckan.params.verbose_extra:
                    print(f"Launching multi-threaded download of multi-file resource {self.name}")
                futures = [executor.submit(self.download_file_query_item_graceful, ckan=ckan, out_dir=out_dir, file_query_item=file_query_item,
                                           index=index, external_stop_event=external_stop_event, start_index=start_index, end_index=end_index, **kwargs)
                           for index, file_query_item in enumerate(self.get_file_query_generator())]
                for future in futures:
                    future.result()  # This will propagate the exception
            total = self.get_file_query_len()
            self._call_progress_callback(total, total, context=f"multi-thread download")
        except Exception as e:
            self.stop_event.set()  # Ensure all threads stop
            if ckan.params.verbose_extra:
                print(f"Stopping all threads because an exception occurred: {e}")
            raise e from e
        finally:
            # self.stop_event.set()  # Ensure all threads stop
            if ckan.params.verbose_extra:
                print("End of multi-threaded download...")
        # at last, apply final actions:
        self._terminate_thread()



#!python3
# -*- coding: utf-8 -*-
"""
Code to initiate a DataStore defined by a large number of files to concatenate into one table.
This concrete implementation is linked to the file system.
"""
from typing import Dict, List, Collection, Callable, Any, Tuple, Generator, Union
import os
from warnings import warn
import glob
import copy

import pandas as pd

from ckanapi_harvesters.auxiliary.error_level_message import ContextErrorLevelMessage, ErrorLevel
from ckanapi_harvesters.builder.mapper_datastore import DataSchemeConversion
from ckanapi_harvesters.builder.builder_errors import ResourceFileNotExistMessage
from ckanapi_harvesters.builder.builder_resource_datastore_multi_abc import BuilderDataStoreMultiABC
from ckanapi_harvesters.builder.builder_resource_datastore_multi_abc import datastore_multi_apply_last_condition_intermediary
from ckanapi_harvesters.auxiliary.ckan_model import UpsertChoice
from ckanapi_harvesters.auxiliary.path import resolve_rel_path, glob_rm_glob, list_files_scandir
from ckanapi_harvesters.ckan_api import CkanApi
from ckanapi_harvesters.auxiliary.ckan_auxiliary import _string_from_element
from ckanapi_harvesters.builder.mapper_datastore_multi import RequestMapperABC, RequestFileMapperABC
from ckanapi_harvesters.builder.mapper_datastore_multi import default_file_mapper_from_primary_key
from ckanapi_harvesters.builder.builder_resource_datastore import BuilderDataStoreFile


class BuilderDataStoreFolder(BuilderDataStoreMultiABC):
    def __init__(self, *, file_query_list: List[Tuple[str,dict]]=None, name:str=None, format:str=None, description:str=None,
                 resource_id:str=None, download_url:str=None, dir_name:str=None):
        super().__init__(name=name, format=format, description=description, resource_id=resource_id, download_url=download_url)
        self.dir_name = dir_name
        # Functions inputs/outputs
        self.df_mapper: RequestFileMapperABC = default_file_mapper_from_primary_key(self.primary_key, file_query_list)
        self.local_file_list_base_dir:Union[str,None] = None
        self.local_file_list:Union[List[str],None] = None
        self.downloaded_file_query_list:Collection[Tuple[str,dict]] = file_query_list

    def copy(self, *, dest=None):
        if dest is None:
            dest = BuilderDataStoreFolder()
        super().copy(dest=dest)
        dest.dir_name = self.dir_name
        dest.local_file_list_base_dir = self.local_file_list_base_dir
        dest.local_file_list = copy.deepcopy(self.local_file_list)
        dest.downloaded_file_query_list = copy.deepcopy(self.downloaded_file_query_list)
        return dest

    def _load_from_df_row(self, row: pd.Series, base_dir:str=None) -> None:
        super()._load_from_df_row(row=row)
        self.df_mapper = default_file_mapper_from_primary_key(self.primary_key)
        self.dir_name: str = _string_from_element(row["file/url"])

    def setup_default_file_mapper(self, primary_key:List[str]=None, file_query_list:Collection[Tuple[str, dict]]=None) -> None:
        """
        This function enables the user to define the primary key and initializes the default file mapper.
        :param primary_key: manually specify the primary key
        :return:
        """
        df_mapper_mem = self.df_mapper
        if primary_key is not None:
            self.primary_key = primary_key
        self.df_mapper = default_file_mapper_from_primary_key(self.primary_key, file_query_list)
        if file_query_list is not None:
            self.downloaded_file_query_list = file_query_list
        # preserve upload/download functions
        self.df_mapper.df_upload_fun = df_mapper_mem.df_upload_fun
        self.df_mapper.df_download_fun = df_mapper_mem.df_download_fun

    @staticmethod
    def resource_mode_str() -> str:
        return "DataStore from Folder"

    def _to_dict(self, include_id:bool=True) -> dict:
        d = super()._to_dict(include_id=include_id)
        d["File/URL"] = self.dir_name
        return d

    @staticmethod
    def from_file_datastore(resource_file: BuilderDataStoreFile,
                            *, dir_name:str=None, primary_key:List[str]=None,
                            file_query_list:Collection[Tuple[str,dict]]=None) -> "BuilderDataStoreFolder":
        resource_folder = BuilderDataStoreFolder()
        resource_folder._load_from_df_row(resource_file._to_row())
        resource_folder.field_builders = resource_file.field_builders
        if dir_name is not None:
            resource_folder.dir_name = dir_name
        elif isinstance(resource_file, BuilderDataStoreFolder):
            resource_folder.dir_name = resource_file.dir_name
        else:
            resource_folder.dir_name, _ = os.path.splitext(resource_file.file_name)
        resource_folder.package_name = resource_file.package_name
        if isinstance(resource_file.df_mapper, RequestMapperABC):
            resource_folder.df_mapper = resource_file.df_mapper.copy()
        else:
            resource_folder.df_mapper.df_upload_fun = resource_file.df_mapper.df_upload_fun
            resource_folder.df_mapper.df_download_fun = resource_file.df_mapper.df_download_fun
        if primary_key is not None or file_query_list is not None:
            resource_folder.setup_default_file_mapper(primary_key=primary_key, file_query_list=file_query_list)
        resource_folder.downloaded_file_query_list = file_query_list
        return resource_folder


    ## upload ---------------------------------------------------
    def upload_file_checks(self, *, resources_base_dir:str=None, ckan: CkanApi=None, **kwargs) -> Union[None,ContextErrorLevelMessage]:
        if os.path.isdir(resolve_rel_path(resources_base_dir, glob_rm_glob(self.dir_name), field=f"File/URL of resource {self.name}")):
            if len(self.list_local_files(resources_base_dir=resources_base_dir)) > 0:
                return None
            else:
                return ResourceFileNotExistMessage(self.name, ErrorLevel.Error,
                    f"Empty resource directory for resource {self.name}: {os.path.join(resources_base_dir, self.dir_name)}")
        else:
            return ResourceFileNotExistMessage(self.name, ErrorLevel.Error,
                f"Missing directory for resource {self.name}: {os.path.join(resources_base_dir, self.dir_name)}")

    def get_sample_file_path(self, resources_base_dir:str, file_index:int=0) -> Union[str,None]:
        self.list_local_files(resources_base_dir=resources_base_dir)
        return self.local_file_list[file_index]

    def load_sample_df(self, resources_base_dir:str, *, upload_alter:bool=True, file_index:int=0, **kwargs) -> pd.DataFrame:
        file_path:str = self.get_sample_file_path(resources_base_dir, file_index=file_index)
        return self.load_local_df(file=file_path, upload_alter=upload_alter, **kwargs)

    def load_local_df(self, file: str, *, upload_alter:bool=True, **kwargs) -> pd.DataFrame:
        # self.sample_data_source = resolve_rel_path(resources_base_dir, self.dir_name, file, field=f"File/URL of resource {self.name}")
        self.sample_data_source = file
        df_local = self.local_file_format.read_file(self.sample_data_source, fields=self._get_fields_info())
        if isinstance(df_local, pd.DataFrame):
            df_local.attrs["source"] = self.sample_data_source
        if upload_alter:
            df_upload = self.df_mapper.df_upload_alter(df_local, self.sample_data_source, fields=self._get_fields_info())
            return df_upload
        else:
            return df_local

    def get_local_file_generator(self, resources_base_dir:str, **kwargs) -> Generator[str, None, None]:
        self.list_local_files(resources_base_dir=resources_base_dir)
        for file_name in self.local_file_list:
            yield file_name

    def get_local_df_generator(self, resources_base_dir:str, **kwargs) -> Generator[pd.DataFrame, None, None]:
        self.list_local_files(resources_base_dir=resources_base_dir)
        for file_name in self.local_file_list:
            yield self.load_local_df(file_name, **kwargs)

    def list_local_files(self, resources_base_dir:str, cancel_if_present:bool=True) -> List[str]:
        if cancel_if_present and self.local_file_list is not None and self.local_file_list_base_dir == resources_base_dir:
            return self.local_file_list
        dir_search_path = resolve_rel_path(resources_base_dir, self.dir_name, field=f"File/URL of resource {self.name}")
        # file_list = [os.path.join(dir_search_path, file_name) for file_name in os.listdir(dir_search_path)]
        # file_list = [os.path.join(file.path, file.name) for file in list(os.scandir(dir_search_path)) if file.is_file()]
        search_query = dir_search_path
        file_list = glob.glob(search_query)
        # file_list = list_files_scandir(dir_search_path)
        file_list.sort()
        self.local_file_list = file_list
        self.local_file_list_base_dir = resources_base_dir
        return file_list

    def init_local_files_list(self, resources_base_dir:str, cancel_if_present:bool=True, **kwargs) -> List[str]:
        return self.list_local_files(resources_base_dir=resources_base_dir, cancel_if_present=cancel_if_present)

    def get_local_file_len(self) -> int:
        if self.local_file_list is None:
            raise RuntimeError("You must call list_local_files first")
        return len(self.local_file_list)

    def upsert_request_df(self, ckan: CkanApi, df_upload:pd.DataFrame,
                          method:UpsertChoice=UpsertChoice.Upsert,
                          apply_last_condition:bool=None, always_last_condition:bool=None) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """
        Call to ckan datastore_upsert.
        Before sending the DataFrame, a call to df_upload_alter is made.
        This implementation optionally checks for the last line of the DataFrame based on the first columns of the primary key.

        :param ckan:
        :param df_upload:
        :param method:
        :return:
        """
        if apply_last_condition is None:
            apply_last_condition = True  # datastore_multi_apply_last_condition_intermediary
        resource_id = self.get_or_query_resource_id(ckan=ckan, error_not_found=True)
        df_upload_local = df_upload
        df_upload_transformed = self.df_mapper.df_upload_alter(df_upload_local, fields=self._get_fields_info())
        file_query = self.df_mapper.get_file_query_of_df(df_upload_transformed)
        if file_query is not None:
            i_restart, upload_needed, row_count, df_row = self.df_mapper.last_inserted_index_request(ckan=ckan,
                                     resource_id=resource_id, df_upload=df_upload_transformed, file_query=file_query)
        else:
            i_restart, upload_needed, row_count, df_row = 0, True, -1, None
        if upload_needed:
            if i_restart > 0 and ckan.params.verbose_extra:
                print(f"Starting transfer from index {i_restart}")
            ret_df = ckan.datastore_upsert(df_upload_transformed.iloc[i_restart:], resource_id, method=method,
                                           apply_last_condition=apply_last_condition,
                                           always_last_condition=always_last_condition, data_cleaner=self.data_cleaner_upload)
        elif 0 <= row_count and row_count < len(df_row):
            msg = f"Sending full dataframe because is was shorter on server side"
            warn(msg)
            ret_df = ckan.datastore_upsert(df_upload_transformed, resource_id, method=method,
                                           apply_last_condition=apply_last_condition,
                                           always_last_condition=always_last_condition, data_cleaner=self.data_cleaner_upload)
        else:
            if ckan.params.verbose_extra:
                print(f"File up to date on server side")
            ret_df = None
        return df_upload_transformed, ret_df


    ## download ---------------------------------------------------------------------------
    def download_file_query_list(self, ckan: CkanApi, cancel_if_present:bool=True) -> List[Tuple[str, dict]]:
        resource_id = self.get_or_query_resource_id(ckan=ckan, error_not_found=self.download_error_not_found)
        if resource_id is None and self.download_error_not_found:
            self.downloaded_file_query_list = []
            return []
        if not(cancel_if_present and self.downloaded_file_query_list is not None):
            file_query_list = self.df_mapper.download_file_query_list(ckan=ckan, resource_id=resource_id)
            self.downloaded_file_query_list = [(self.df_mapper.get_file_name_of_query(file_query), file_query) for file_query in file_query_list]
        return self.downloaded_file_query_list

    def setup_download_file_query_list(self, file_query_list: List[Tuple[str,dict]]) -> None:
        self.downloaded_file_query_list = file_query_list

    def init_download_file_query_list(self, ckan: CkanApi, out_dir: str, cancel_if_present:bool=True, **kwargs) -> List[Any]:
        if out_dir is not None:
            dir_tables = resolve_rel_path(out_dir, glob_rm_glob(self.dir_name, default_rec_dir=self.name), field=f"File/URL of resource {self.name}")
            os.makedirs(dir_tables, exist_ok=True)
        return self.download_file_query_list(ckan=ckan, cancel_if_present=cancel_if_present)

    def get_file_query_len(self) -> int:
        if self.downloaded_file_query_list is None:
            raise RuntimeError("You must call download_file_query_list first")
        return len(self.downloaded_file_query_list)

    def get_file_query_generator(self) -> Generator[Tuple[str,dict], Any, None]:
        for file_name, file_query in self.downloaded_file_query_list:
            yield file_name, file_query

    def download_file_query(self, ckan: CkanApi, out_dir: str, file_name:str, file_query:dict) \
            -> Tuple[Union[str,None], Union[pd.DataFrame,None]]:
        resource_id = self.get_or_query_resource_id(ckan=ckan, error_not_found=self.download_error_not_found)
        if resource_id is None and self.download_error_not_found:
            return None, None
        self.download_file_query_list(ckan, cancel_if_present=True)
        file_out = None
        if out_dir is not None:
            file_out = resolve_rel_path(out_dir, glob_rm_glob(self.dir_name, default_rec_dir=self.name), file_name, field=f"File/URL of resource {self.name}")
            if self.download_skip_existing and os.path.exists(file_out):
                if ckan.params.verbose_extra:
                    print(f"Skipping existing file {file_out}")
                return file_out, None
        df_download = self.df_mapper.download_file_query(ckan=ckan, resource_id=resource_id, file_query=file_query)
        df = self.df_mapper.df_download_alter(df_download, file_query=file_query, fields=self._get_fields_info())
        if out_dir is not None:
            self.local_file_format.write_file(df, file_out, fields=self._get_fields_info())
        else:
            file_out = None
        return file_out, df

    def download_file_query_item(self, ckan: CkanApi, out_dir: str, file_query_item: Tuple[str,dict]) -> Tuple[str, pd.DataFrame]:
        file_name, file_query = file_query_item
        return self.download_file_query(ckan=ckan, file_name=file_name, file_query=file_query, out_dir=out_dir)

    def download_request(self, ckan: CkanApi, out_dir: str, *, full_download:bool=False, force:bool=False, threads:int=1) -> None:
        # limit download to first page by default
        if not full_download:
            super().download_request(ckan=ckan, out_dir=out_dir, full_download=False, force=force, threads=threads)
        else:
            self.download_request_full(ckan=ckan, out_dir=out_dir, threads=threads, force=force)


